/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_QuadTree_h__
#define __surface_QuadTree_h__

namespace fe
{

/**************************************************************************//**
	@brief Triangular storage using divisions by quarters

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT QuadTree: public FlatTree, public ObjectSafeShared<QuadTree>
{
	public:
				QuadTree(void);
virtual			~QuadTree(void)												{}

virtual	void	setRefinement(U32 a_refinement)	{ m_subdivMax=a_refinement; }

virtual	void	populate(const SpatialVector* a_pVertexArray,
						const SpatialVector* a_pNormalArray,
						U32 a_vertices,
						const SpatialVector& a_center,F32 a_radius);
virtual	Real	nearestPoint(const SpatialVector& a_origin,U32 a_hitLimit,
						std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray)
						const;
virtual	Real	rayImpact(const SpatialVector& a_origin,
						const SpatialVector& a_direction,Real a_maxDistance,
						U32 a_hitLimit,
						std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray)
						const;

virtual	void	draw(sp<DrawI> a_spDrawI,const fe::Color* a_pColor) const;

		Real	nearestTo(BWORD a_mutable,BWORD a_directional,
						const SpatialVector& a_origin,
						const SpatialVector& a_direction,
						Real a_maxDistance,U32 a_hitLimit,
						std::vector< sp<SpatialTreeI::Hit> >&
						a_rHitArray,U32 a_subdivMax) const;

		void	age(void);

	class Element
	{
		public:
					Element(void):
						m_index(0),
						m_sphereIndex(0),
						m_sortValue(0.0)									{}
			U32		m_index;
			U32		m_sphereIndex;
			F32		m_sortValue;

	static	int		compare(const void* a_pVoid1,const void* a_pVoid2);
	};

	class FE_DL_EXPORT BaseNode: public DAGNode
	{
		public:
			U32				m_level;
			Vector4			m_boundSphere;
			String			m_debugText;

							BaseNode(void):
								m_level(0)
							{	set(m_boundSphere); }

							~BaseNode(void)									{}

			void			subdivide(const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Element* a_pElements,
									U32 a_elementIndex,U32 a_elementCount,
									U32 a_grain,U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp);

	virtual	void			subdivideOne(const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									U32 a_triIndex,U32 a_grain,
									U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp)						=0;
	virtual	void			subdivideAll(const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									U32 a_grain,U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp)						=0;

			void			draw(sp<DrawI> a_spDrawI,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									U32 a_grain,const Color* a_pColor,
									U32 a_level) const;
	static	void			calcFrames(
									SpatialTransform& a_rTransform1,
									SpatialTransform& a_rTransform2,
									const SpatialVector& a_rVertex1,
									const SpatialVector& a_rNormal1,
									const SpatialVector& a_rVertex2,
									const SpatialVector& a_rNormal2);
			Real			separation(const SpatialVector& a_origin) const;

			void			nearestTo(BWORD a_mutable,BWORD a_directional,
									const SpatialVector& a_origin,
									const SpatialVector& a_direction,
									Real a_maxDistance,U32 a_hitLimit,
									std::vector< sp<SpatialTreeI::Hit> >&
									a_rHitArray,
									CountedPool<Hit>& a_rHitPool,
									U32 a_subdivMax,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									U32 a_grain,const Vector4* a_pSphereArray,
									U32 a_timeStamp) const;

	virtual	U32				triangleCount(void) const						=0;
	virtual
	const SpatialVector*	triangleVertex(U32 index,const SpatialVector*
									a_pVertexArray) const					=0;
	virtual
	const SpatialVector*	triangleNormal(U32 index,const SpatialVector*
									a_pNormalArray) const					=0;
	virtual
	const Vector4			sphereData(U32 index,
									const Vector4* a_pSphereArray) const	=0;

			void			prune(U32 a_timeStamp);
	};
	class FE_DL_EXPORT Node: public BaseNode
	{
		public:
			Element*		m_pElements;
			U32				m_elementCount;

							Node(void):
								m_pElements(NULL),
								m_elementCount(0)							{}

							~Node(void)
							{	delete[] m_pElements; }

			void			refine(const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									SpatialVector* a_pVertexScratch,
									U32 a_grain,const Vector3u& a_divisions,
									const Vector4* a_pSphereArray,
									U32 a_capacity);
			void			split(const Vector4* a_pSphereArray,U32 axis,
									Element*& a_rpElements,
									U32& a_rElementCount,U32 divisions,
									Element** a_rppElements,
									U32* a_rpElementCount);
			void			calcBounds(const SpatialVector* a_pVertexArray,
									SpatialVector* a_pVertexScratch,
									U32 a_grain);
			void			calcBounds(const SpatialVector* a_pVertexArray,
									U32 a_grain);

	virtual	void			subdivideOne(const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									U32 a_triIndex,U32 a_grain,
									U32 a_level,U32 a_maxLevel,U32 a_timeStamp);
	virtual	void			subdivideAll(const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									U32 a_grain,U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp);

	virtual	U32				triangleCount(void) const
							{	return m_elementCount; }
	virtual
	const SpatialVector*	triangleVertex(U32 index,
									const SpatialVector* a_pVertexArray) const
							{	FEASSERT(index<m_elementCount);
								return &a_pVertexArray
										[m_pElements[index].m_index*3]; }
	virtual
	const SpatialVector*	triangleNormal(U32 index,
									const SpatialVector* a_pNormalArray) const
							{	FEASSERT(index<m_elementCount);
								return &a_pNormalArray
										[m_pElements[index].m_index*3]; }
	virtual
	const Vector4			sphereData(U32 index,
									const Vector4* a_pSphereArray) const
							{	FEASSERT(index<m_elementCount);
//								feLog("QuadTree::BaseNode::sphereData"
//										" elements %p index %d\n",
//										m_pElements,
//										m_pElements[index].m_sphereIndex);
								return a_pSphereArray
										[m_pElements[index].m_sphereIndex]; }
	};
	class FE_DL_EXPORT SubNode: public BaseNode
	{
		public:
			SpatialVector*	m_pSubdivVertexArray;
			SpatialVector*	m_pSubdivNormalArray;
			U32				m_serial;
			U32				m_subdivCount;
			U32				m_subdivLevel;
			U32				m_timeStamp;

							SubNode(void):
								m_pSubdivVertexArray(NULL),
								m_pSubdivNormalArray(NULL),
								m_serial(0),
								m_subdivCount(0),
								m_subdivLevel(0),
								m_timeStamp(0)								{}

							~SubNode(void)
							{
								delete[] m_pSubdivVertexArray;
								delete[] m_pSubdivNormalArray;
							}

	virtual	void			subdivideOne(const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									U32 a_triIndex,U32 a_grain,
									U32 a_level,U32 a_maxLevel,U32 a_timeStamp);
	virtual	void			subdivideAll(const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									U32 a_grain,U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp);

	virtual	U32				triangleCount(void) const
							{	return m_subdivCount/3; }
	virtual
	const SpatialVector*	triangleVertex(U32 index,
									const SpatialVector* a_pVertexArray) const
							{	FEASSERT(index*3+2<m_subdivCount);
								return &m_pSubdivVertexArray[index*3]; }
	virtual
	const SpatialVector*	triangleNormal(U32 index,
									const SpatialVector* a_pNormalArray) const
							{	FEASSERT(index*3+2<m_subdivCount);
								return &m_pSubdivNormalArray[index*3]; }
	virtual
	const Vector4			sphereData(U32 index,
									const Vector4* a_pSphereArray) const
							{	return Vector4(0.0f,0.0f,0.0f,0.0f); }
	};

	class Child
	{
		public:
							Child(void): m_spNode(NULL),m_distance(0.0)		{}
			sp<BaseNode>	m_spNode;
			F32				m_distance;

	static	int			compare(const void* a_pVoid1,const void* a_pVoid2);
	};

		void			cacheSpheres(const SpatialVector& a_center,
								F32 a_radius);
		void			buildGraph(void);

		sp<Node>		m_spRootNode;
		sp<DrawMode>	m_spWireframe;

	protected:
		void				prune(void);

		Vector3u			m_divisions;
		U32					m_capacity;
		U32					m_subdivMax;
		U32					m_ageStamp;
		U32					m_timeStamp;
};

} /* namespace */

#endif /* __surface_QuadTree_h__ */

